import numpy as np
from .find_zero_eigenvalue_eigenvector import find_zero_eigenvalue_eigenvector

### Private functions

# Scale Q so that the average mutation rate is 0.01
def scale_q(Q, EQ):
        SCALE_FACTOR = np.dot(EQ, (-np.diag(Q)))

        if(SCALE_FACTOR == 0):
            raise ZeroDivisionError('No Q diagonal cause a problem in estimate_q.py:scale_q')

        Q /= SCALE_FACTOR

        return Q

#   Sometimes, when data is sparse, Q estimates come out with
#   off-diagonal entries being negative. Not good.
def fix_negatives(Q):
    # Replace negative elements with smallest absolute value
    MINIMUM_ELEMENT = np.min(np.abs(Q))
    Q[Q<0] = MINIMUM_ELEMENT
    #   Recalculate diagonal to be negative rowsums
    np.fill_diagonal(Q, 0)
    ROW_SUMS = Q.sum(axis=1)
    np.fill_diagonal(Q, -ROW_SUMS)

    return Q

def recover_q(L, VR, VL):
    Q = 0.01 * (VR @ np.diag(L) @ VL)
    return Q

# Alternative: Estimate eigenvalues of Q using weighted points
#
# The estimated eigenvalues are returned in L
def _weighted_estimate_eigenvals(PW, W, VL, VR, DIST_SAMPLES):
    # The X and Y matrises will contain 20 least-squares problems, one for each eigenvalue
    NUMBER_OF_DIST_SAMPLES = len(DIST_SAMPLES)
    X = np.empty(NUMBER_OF_DIST_SAMPLES, dtype="float64").reshape((80,1))
    Y = np.empty((20, NUMBER_OF_DIST_SAMPLES))

    # Find the eigenvector corresponding to eigenvalue = 1
    _, NULL_VECTOR_INDEX = find_zero_eigenvalue_eigenvector(VL)

    # Gather some datapoints
    for i, DIST_SAMPLE in enumerate(DIST_SAMPLES):
        ELAMBDA = np.diag(VL @ PW[i] @ VR)
        X[i] = (DIST_SAMPLE / 100) * W[i]

        for li in range(20):
            if (li == NULL_VECTOR_INDEX):
                continue

            if (ELAMBDA[li] > 0):    # Skip complex value data points!
                Y[li, i] = np.real(np.log(ELAMBDA[li])) * W[i]
            else:
                X[i] = 0   # No disturbance if set to 0!
                Y[li, i] = 0

    L = np.zeros(20)

    for i,_ in enumerate(L):
        if(i == NULL_VECTOR_INDEX):
            L[i] = 0
        else:
            tempY = Y[i,:].reshape(80,1)
            L[i] = np.linalg.lstsq(X, tempY, rcond = None)[0]

    return L


### Interface
def estimate_q(PW, W, VL, VR, EQ, DIST_SAMPLES):
    L = _weighted_estimate_eigenvals(PW, W, VL, VR, DIST_SAMPLES)

    Q = recover_q(L, VR, VL)
    Q = fix_negatives(Q)
    Q = scale_q(Q, EQ)

    return Q



__version__ = '2.1.3'

### Version history
#
# 2.0   - Ruben Riddeström's BSc thesis work
# 2.1   - Lars' adaptions
# 2.1.1 - Updated README, corrected URL,
# 2.1.2 - Fixed content type on README for PyPi
# 2.1.3 - Updated unit tests and corrected Octave output



import sys
import os
import io
from contextlib import redirect_stdout
import numpy as np
from modelestimator.main import main

#
# Verify Octave-style output. This test is probably too strict. Should be connected to an Octave process, but cannot
# really demand that from the user.
#

def test_integration_2():
    #   Load REFERENCE_OUTPUT_STRING
    reference_output_string_path = "tests/test_integration_2/reference.txt"
    reference_output_string_path = os.path.join(sys.path[0], reference_output_string_path)

    with open(reference_output_string_path) as text_file:
            REFERENCE_OUTPUT_STRING = text_file.read()

    #   Calculate OUTPUT_STRING
    reference_input_string_path = "/tests/test_integration_2/test_integration_1_20seqs_1000long_50pam.fa"
    reference_input_string_path = sys.path[0] + reference_input_string_path
    sys.argv = ["-", "-f", "fasta", "-a", "octave", reference_input_string_path]

    f = io.StringIO()
    with redirect_stdout(f):
        main()

    OUTPUT_STRING = f.getvalue()

    assert(OUTPUT_STRING == REFERENCE_OUTPUT_STRING)



import numpy as np
from scipy.linalg import eig
from .find_zero_eigenvalue_eigenvector import find_zero_eigenvalue_eigenvector

def find_eigens(COUNT_MATRIX_LIST):
    P_SUM = sum(0.5 * (MATRIX + MATRIX.T) for MATRIX in COUNT_MATRIX_LIST)

    # Make every row sum to 1
    ROW_SUMS = np.linalg.norm(P_SUM, axis=1, ord=1, keepdims=1)
    P_SUM = np.divide(P_SUM, ROW_SUMS, out=np.zeros_like(P_SUM), where=ROW_SUMS != 0)   # Only divide where the row sum is non-zero

    try:
        EIGEN_VALUES, VR = eig(P_SUM, left=False, right=True)   #   Calculate eigenvalues and the right eigenvectors of P_SUM
    except ValueError:
        raise ValueError("Unable to calculate eigenvalues")

    if not np.all(np.isreal(EIGEN_VALUES)):
        raise ValueError("An eigenvalue is complex")

    VL = np.linalg.inv(VR)

    EQ,_ = find_zero_eigenvalue_eigenvector(VL)
    EQ /= np.linalg.norm(EQ, ord=1)
    EQ = np.absolute(EQ)

    return VL, VR, EQ


import argparse
import sys

from modelestimator.version import __version__

from modelestimator._bw_estimator.bw_estimator import bw_estimator
from modelestimator.io import handle_input_file, format_model_output
from modelestimator.bootstrap import bootstrapped_stability, q_bootstrap_estimate



description='''
Infer a amino-acid replacement-rate matrix from one or more protein multisequence
alignments. The output is a rate matrix and an associated steady-state amino-acid
distribution vector.
'''
example_usage='''
Example usage:
    modelestimator -t 0.001 file1.fa file2.fa file3.fa
    modelestimator -b 200 file.fa
'''

def setup_argument_parsing():
    '''
    Create an argument parser.
    '''
    ap = argparse.ArgumentParser(description=description, epilog=example_usage,
                                 formatter_class=argparse.RawDescriptionHelpFormatter)
    ap.add_argument('infiles', nargs='+', help="One or more infiles, containing protein multialignments in FASTA format or chosen according to the -f/--format option.")
    ap.add_argument('--version', action='version', version='%(prog)s ' + __version__)
    ap.add_argument('-a', '--application', choices=['iqtree', 'mrbayes', 'octave', 'paml', 'phyml', 'raxml'], default='paml',
                    help="Choose output format to suit the application you want to use for inference. The 'iqtree', 'paml', 'phyml', and 'raxml' options are identical. The 'octave' option is for import into Octave (a MatLab-compatible program) and is presenting the actual Q matrix rather than the R matrix used by PAML/PhyML, etc. Default: %(default)s")
    ap.add_argument('-f', '--format', choices=['fasta', 'clustal', 'nexus', 'phylip', 'stockholm'], default='fasta',
                    help="Specify sequence format of input files. Default: %(default)s")
    # ap.add_argument('-f', '--format', choices=['guess', 'fasta', 'clustal', 'nexus', 'phylip', 'stockholm'], default='guess',
    #                 help="Specify what sequence type to assume. Be specific if the file is not recognized automatically. When reading from stdin, the format is always guessed to be FASTA. Default: %(default)s")
    ap.add_argument('-t', '--threshold', type=float, metavar='T', default=0.001,
                    help="Stop when consecutive iterations do not change by more than T. Default: %(default)g")
    ap.add_argument('-b', '--bootstrap', type=int, metavar='N', default=None,
                    help="Estimate the rate matrix using bootstrapping by computing  N resampled replicates of the input multialignment. For each replicate, a rate matrix is computed. The mean matrix the elementwise standard deviation is returned. Only one infile should be given in this mode.")
    ap.add_argument('-B', '--bootstrapped_quality', type=int, metavar='N', default=None,
                    help="Estimate the quality of the rate matrix estimate using a bootstrap procedure. The multialignment is resampled N times and a Q matrix is computed for each replicate. Then the difference (matrix norm) between rate matrix estimated without resampling and each bootstrapped Q is computed and the mean difference is returned. Only one infile should be given in this mode. Returns bootstrap norm.")
    return ap


def main():
    ap = setup_argument_parsing()
    args = ap.parse_args()

    multialignment_list = []

    for f in args.infiles:
        try:
            multialignment = handle_input_file(f, args.format)
            multialignment_list.append(multialignment)
        except Exception as e:
            print(f'Error reading alignments, file "{f}":', e, file=sys.stderr)
            sys.exit(1)
    try:
        if args.bootstrapped_quality:
            multialignment = multialignment_list[0]
            bootstrap_norm, _ = bootstrapped_stability(args.bootstrapped_quality, args.threshold, multialignment)
            print('Bootstrap norm = ' + str(bootstrap_norm))

        elif args.bootstrap:
            Q, eq, Q_sd, n_failures = q_bootstrap_estimate(multialignment_list, args.threshold, args.bootstrap)
            print('This is, so far, nonsensical output. The R matrix is correct, but it is unclear what the SD matrix is. It is treated as a Q matrix.')
            output_string = format_model_output(Q, eq, args.application)
            output_string += '\n\n' + format_model_output(Q_sd, eq, args.application)
            print(output_string)
            if n_failures > 0:
                print(f'Warning: the bootstrap failed in {n_failures} replicates.')

        else:
            Q, EQ = bw_estimator(args.threshold, multialignment_list)
            output_string = format_model_output(Q, EQ, args.application)
            print(output_string)

    except Exception as e:
        print('Error:', e, file=sys.stderr)
        sys.exit(1)



import os
import sys
import numpy as np
from .comp_posterior_JC import comp_posterior_JC
from .matrix_weight import matrix_weight
from .estimate_q import estimate_q
from .simple_estimation import simple_estimation
from .find_eigens import find_eigens

def calculate_q_eq(COUNT_MATRIX_LIST, THRESHOLD):
    #   VL = Inverse of right eigenvectorts
    #   VR = Right eigenvectors
    #   EQ = Right eigenvector corresponding to zero eigenvalue but normalized so it sums to 1
    VL, VR, EQ = find_eigens(COUNT_MATRIX_LIST)

    #   Get a first simple estimate of Q using a Jukes-Cantor model
    DIST_SAMPLES = np.arange(1, 400, 5)
    POSTERIOR = comp_posterior_JC(COUNT_MATRIX_LIST, DIST_SAMPLES)   # posterior.shape = (10, 80). Rows are identical to Octave but in different order
    PW = matrix_weight(COUNT_MATRIX_LIST, POSTERIOR, DIST_SAMPLES)    
    W = POSTERIOR.sum(axis=0)
    q = estimate_q(PW, W, VL, VR, EQ, DIST_SAMPLES)

    #   Set loop variables
    difference = 1+THRESHOLD
    iterations = 0
    MAX_ITERATIONS = 10
    
    #   Calculate Q
    while (iterations < MAX_ITERATIONS and difference > THRESHOLD):
        iterations += 1
        q_new = simple_estimation(COUNT_MATRIX_LIST, q, VL, VR, EQ, DIST_SAMPLES)
        difference = np.linalg.norm(q_new - q)
        q = q_new
    return q, EQ



import numpy as np

### Interface
# Weighting the count matrices.
# PW is a list, with the sum of count matrices weighted by their posterior probability
# for evolutionary distance D.
def matrix_weight(COUNT_MATRIX_LIST, POSTERIOR, DIST_SAMPLES):
    NUMBER_OF_DIST_SAMPLES = len(DIST_SAMPLES)
    PW = np.empty((NUMBER_OF_DIST_SAMPLES, 20, 20))
    
    PW = np.tensordot(POSTERIOR, COUNT_MATRIX_LIST, axes=([0],[0]))
    with np.errstate(invalid='ignore'):  # Suppress warning for division with NaN
        PW /= PW.sum(axis=2,keepdims=True)  # Normalize matrix rows
        
    return PW


import numpy as np

#   Find the index of the eigenvector corresponding to Q's zero eigenvalue.
#   This is recognized as the row (because we will be looking at the 'right'
#   eigenvectors, not the usual left) with all positive or all negative elements.
def find_zero_eigenvalue_eigenvector(MATRIX):
    ZERO_EIGEN_VECTOR_ARRAY = np.array([(EIGEN_VECTOR, INDEX) for INDEX, EIGEN_VECTOR in enumerate(MATRIX) if all(EIGEN_VECTOR > 0) or all(EIGEN_VECTOR < 0)])

    if (len(ZERO_EIGEN_VECTOR_ARRAY) > 1):
        raise ValueError("More than one candidate for null-vector")
    if (len(ZERO_EIGEN_VECTOR_ARRAY) == 0):
        raise ValueError("No candidate for null-vector!")
        
    EIGEN_VECTOR, INDEX = ZERO_EIGEN_VECTOR_ARRAY[0]

    #   Returns as list to return copy and not reference
    return list(EIGEN_VECTOR), INDEX


import tempfile
import numpy as np
import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from modelestimator._bw_estimator.bw_estimator import bw_estimator
from modelestimator.io import handle_input_file

def test_case_3(tmpdir):
        #   Create directory paths
    CURRENT_DIR = os.path.dirname(__file__)
    TEST_FILES_PATH = os.path.join(CURRENT_DIR, 'test_case_3/')

    #   Create sequence file path
    FILE_PATH_1 = os.path.join(TEST_FILES_PATH, "JTT_balancedtree_32sequences_10000long_1.fa")
    FILE_PATH_2 = os.path.join(TEST_FILES_PATH, "JTT_balancedtree_32sequences_10000long_2.fa")
    FILE_PATH_3 = os.path.join(TEST_FILES_PATH, "JTT_balancedtree_32sequences_10000long_3.fa")

    #   Load reference Q and EQ
    REFERENCE_Q_PATH = os.path.join(TEST_FILES_PATH, 'test_case_3_Q.txt')
    REFERENCE_Q = np.loadtxt(REFERENCE_Q_PATH)
    REFERENCE_EQ_PATH = os.path.join(TEST_FILES_PATH, 'test_case_3_EQ.txt')
    REFERENCE_EQ = np.loadtxt(REFERENCE_EQ_PATH)

    #   Calculate Q and EQ
    FILE_PATH_LIST = [FILE_PATH_1, FILE_PATH_2, FILE_PATH_3]
    FORMAT = "fasta"
    MULTIALIGNMENT_LIST = []
    for FILE in FILE_PATH_LIST:
        MULTIALIGNMENT = handle_input_file(FILE, FORMAT)
        MULTIALIGNMENT_LIST.append(MULTIALIGNMENT)
    THRESHOLD = 0.001
    CALCULATED_Q, CALCULATED_EQ = bw_estimator(THRESHOLD, MULTIALIGNMENT_LIST)

    #   Assert that calculated and references are close. Expected to pass
    print(CALCULATED_Q)
    assert(np.allclose(CALCULATED_Q, REFERENCE_Q, atol=THRESHOLD))
    assert(np.allclose(CALCULATED_EQ, REFERENCE_EQ, atol=THRESHOLD))



# coding=utf-8

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('modelestimator/version.py') as fh: # __version__ is defined here
    exec(fh.read())

setuptools.setup(
	name='modelestimator-v2',
	version=__version__,
	author='Ruben Ridderström and Lars Arvestad',
	author_email='ruben.ridderstrom@gmail.com',
	description='Program for estimating amino acid replacement rates',
	long_description=long_description,
        long_description_content_type="text/markdown; charset=UTF-8",
	url='https://github.com/arvestad/modelestimator-v2',
	license='GPLv3',
	packages = setuptools.find_packages(),
	entry_points={
		'console_scripts':[
			'modelestimator = modelestimator.main:main'
		]
	},
	install_requires=[
            'argparse',
	    "scipy",
	    "numpy",
	    "biopython"
	],
	setup_requires=['pytest-runner'],
	tests_require=['pytest']
	)



import numpy as np

### Private functions
def _matching_letters(a,b, COMPARE_INDELS_FLAG):
    if (len(a) != len(b)):
        raise ValueError("Sequences need to be of equal length")

    number_of_matching_positions = 0

    if COMPARE_INDELS_FLAG:
        for x,y in zip(a,b):
            if x == y and x != '-' :
                number_of_matching_positions += 1
    else:
        number_of_matching_positions = np.sum(a==b) # Faster than when indels have to be ignored
                   
    return number_of_matching_positions

### Interface
def choose_close_pairs(sequence_list, COMPARE_INDELS_FLAG):
    '''
    Greedily choose a list of sequence pairs based on sequence identity.
    The objective is to avoid the furthermost pairs and to avoid 
    over-sampling of some sequences and edges (in the underlying tree).

    Returns a list of sequence pairs.
    '''
    indexes_and_matching_letters = []

    for PRIMARY_INDEX, PRIMARY_SEQUENCE in enumerate(sequence_list):
        for SECONDARY_INDEX in range(PRIMARY_INDEX+1, len(sequence_list)):
            SECONDARY_SEQUENCE = sequence_list[SECONDARY_INDEX]
            MATCHING_LETTERS = _matching_letters(PRIMARY_SEQUENCE, SECONDARY_SEQUENCE, COMPARE_INDELS_FLAG)

            INDEXES = (PRIMARY_INDEX, SECONDARY_INDEX)
            INDEX_SCORE_TUPLE = (INDEXES, MATCHING_LETTERS)
            indexes_and_matching_letters.append(INDEX_SCORE_TUPLE)

    indexes_and_matching_letters.sort(key=lambda tup: tup[1], reverse = True)   # Sort on matching letters

    matched_indexes = []
    close_pairs = []
    NUMBER_OF_SEQUENCES = len(sequence_list)

    while (NUMBER_OF_SEQUENCES - len(matched_indexes)) >= 2:
        CURRENT_INDEX_AND_MATCHING_LETTERS_TUPLE = indexes_and_matching_letters.pop(0)
        index1, index2 = CURRENT_INDEX_AND_MATCHING_LETTERS_TUPLE[0]

        if not(index1 in matched_indexes) and not(index2 in matched_indexes):
            matched_indexes.append(index1)
            matched_indexes.append(index2)

            seq1= sequence_list[index1]
            seq2 = sequence_list[index2]
            close_pairs.append((seq1, seq2))

    return close_pairs
